from typing import Union

from pika import BasicProperties
from pika.adapters.blocking_connection import BlockingChannel

from petisco.base.domain.message.message import Message
from petisco.extra.rabbitmq.application.message.formatter.rabbitmq_message_queue_name_formatter import (
    RabbitMqMessageQueueNameFormatter,
)


class RabbitMqMessagePublisher:
    def __init__(self, exchange_name: str) -> None:
        self._exchange_name = exchange_name
        self._properties = BasicProperties(delivery_mode=2)  # PERSISTENT_TEXT_PLAIN

    def execute(
        self,
        channel: BlockingChannel,
        message: Message,
        routing_key: Union[str, None] = None,
        first_time: bool = False,
    ) -> None:
        if routing_key is None:
            routing_key = RabbitMqMessageQueueNameFormatter.format(message, exchange_name=self._exchange_name)

        if first_time:
            # Confirm delivery should be enabled just once. Source: https://www.rabbitmq.com/tutorials/tutorial-seven-java#enabling-publisher-confirms-on-a-channel
            # Otherwise with Pika we get lots of error messages
            channel.confirm_delivery()
        channel.basic_publish(
            exchange=self._exchange_name,
            routing_key=routing_key,
            body=message.format_json().encode(),
            properties=self._properties,
        )



from typing import List, Union

from pika.exceptions import ChannelClosedByBroker

from petisco.base.application.chaos.check_chaos import check_chaos_publication
from petisco.base.domain.message.command import Command
from petisco.base.domain.message.command_bus import CommandBus
from petisco.extra.rabbitmq.application.message.bus.rabbitmq_message_publisher import (
    RabbitMqMessagePublisher,
)
from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_configurer import (
    RabbitMqMessageConfigurer,
)
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_consumer_connector import (
    RabbitMqConsumerConnector,
)
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector


class RabbitMqCommandBus(CommandBus):
    """
    An implementation of CommandBus using RabbitMQ infrastructure.
    Implementation is based on pika library.
    """

    def __init__(
        self,
        organization: str,
        service: str,
        connector: Union[RabbitMqConnector, RabbitMqConsumerConnector] = RabbitMqConnector(),
        fallback: Union[CommandBus, None] = None,
    ):
        self.connector = connector
        self.exchange_name = f"{organization}.{service}"
        self.rabbitmq_key = f"publisher-{self.exchange_name}"
        self.configurer = RabbitMqMessageConfigurer(organization, service, connector)
        self.already_configured = False
        self.fallback = fallback
        self.publisher = RabbitMqMessagePublisher(self.exchange_name)

    def dispatch(self, command: Union[Command, List[Command]]) -> None:
        """
        Dispatch one Command or a list of commands

        Dispatch several commands could be a code smell but some use case could require this feature.
        """

        meta = self.get_configured_meta()
        dispatched_commands = []
        commands = self._check_input(command)

        try:
            check_chaos_publication()
            channel = self.connector.get_channel(self.rabbitmq_key)
            for i, command in enumerate(commands):
                self._check_is_command(command)
                command = command.update_meta(meta)
                self.publisher.execute(channel, command, first_time=i == 0)
                if channel.is_open and not isinstance(self.connector, RabbitMqConsumerConnector):
                    channel.close()
                dispatched_commands.append(command)
        except ChannelClosedByBroker:
            unpublished_commands = [command for command in commands if command not in dispatched_commands]
            self._retry(unpublished_commands)
        except Exception as exc:  # noqa
            if not self.fallback:
                raise exc

            unpublished_commands = [command for command in commands if command not in dispatched_commands]
            self.fallback.dispatch(unpublished_commands)

    def _retry(self, command: Union[Command, List[Command]]) -> None:
        # If command queue is not configured, it will be configured and then try to dispatch again.
        if not self.already_configured:
            self.configurer.configure()
            self.already_configured = True
            self.dispatch(command)
        elif self.fallback:
            self.fallback.dispatch(command)

    def close(self) -> None:
        """
        Close RabbitMQ connection.
        """
        self.connector.close(self.rabbitmq_key)



import logging
import os
from typing import Callable

from petisco.extra.logger.log_message import LogMessage
from petisco.extra.logger.logger import (
    CRITICAL,
    DEBUG,
    ERROR,
    FATAL,
    INFO,
    WARN,
    WARNING,
    Logger,
)

CORRESPONDENCES_LOGGING_LEVEL = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


class LoggingBasedLogger(Logger):
    def __init__(
        self,
        logger_name: str,
        format: str = "%(name)s - %(levelname)s - %(message)s",
        config: Callable = None,
    ):
        self.logging_level = os.environ.get("PETISCO_LOGGING_LEVEL", "INFO").upper()
        logging_level_value = CORRESPONDENCES_LOGGING_LEVEL.get(self.logging_level, logging.INFO)
        logging.basicConfig(format=format, level=logging_level_value)
        if config:
            config()
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(logging_level_value)
        self.logger.info(f"Set PETISCO_LOGGING_LEVEL: {self.logging_level}")

    def log(self, logging_level, log_message: LogMessage):
        message = log_message.to_dict()
        if logging_level == INFO:
            message["meta"]["level"] = "info"
            self.logger.info(message)
        elif logging_level == DEBUG:
            message["meta"]["level"] = "debug"
            self.logger.debug(message)
        elif logging_level == CRITICAL:
            message["meta"]["level"] = "critical"
            self.logger.critical(message)
        elif logging_level == FATAL:
            message["meta"]["level"] = "fatal"
            self.logger.fatal(message)
        elif logging_level == ERROR:
            message["meta"]["level"] = "error"
            self.logger.error(message)
        elif logging_level == WARN:
            message["meta"]["level"] = "warning"
            self.logger.warn(message)
        elif logging_level == WARNING:
            message["meta"]["level"] = "warning"
            self.logger.warning(message)



from typing import Any, Union

from meiga import BoolResult, Error, Failure, isSuccess

from petisco.base.application.patterns.repository import Repository
from petisco.base.domain.errors.defaults.already_exists import (
    AggregateAlreadyExistError,
)
from petisco.base.domain.errors.defaults.not_found import (
    AggregateNotFoundError,
    AggregatesNotFoundError,
)
from petisco.base.domain.model.uuid import Uuid
from petisco.base.domain.persistence.databases import databases
from petisco.extra.elastic.elastic_database import ElasticDatabase, ElasticSessionScope


class ElasticRepository(Repository):
    session_scope: ElasticSessionScope

    def __init__(self, database_alias: str):
        """
        Constructs the ElasticRepository from a database_alias creating a session_scope (ElasticSessionScope)

        Args:
            database_alias (str): A str you defined in database_alias the DatabaseConfigurer
        """
        database = databases.get(ElasticDatabase, alias=database_alias)
        self.session_scope = database.get_session_scope()

    @classmethod
    def fail_if_aggregate_already_exist(
        cls, model: Any, aggregate_id: Uuid, result_error: Union[Error, None] = None
    ) -> BoolResult:
        if model:
            error = (
                result_error
                if result_error
                else AggregateAlreadyExistError(aggregate_id, cls.__name__, model.__tablename__)
            )
            return Failure(error)
        return isSuccess

    @classmethod
    def fail_if_aggregate_not_found(
        cls, model: Any, aggregate_id: Uuid, result_error: Union[Error, None] = None
    ) -> BoolResult:
        if not model:
            error = result_error if result_error else AggregateNotFoundError(aggregate_id, cls.__name__)
            return Failure(error)
        return isSuccess

    @classmethod
    def fail_if_aggregates_not_found(cls, model: Any, result_error: Union[Error, None] = None) -> BoolResult:
        if not model:
            error = result_error if result_error else AggregatesNotFoundError(cls.__name__)
            return Failure(error)
        return isSuccess



from contextlib import contextmanager
from typing import Callable, ContextManager, Iterator

from loguru import logger
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session

from petisco.extra.meiga import WaitingForEarlyReturn


def sql_session_scope_provider(
    session_factory: Callable[[], Session],
) -> Callable[[], ContextManager[Session]]:
    @contextmanager
    def session_scope() -> Iterator[Session]:
        session = session_factory()
        try:
            yield session
            session.commit()
        except WaitingForEarlyReturn as e:
            session.rollback()
            raise e
        except (OperationalError, Exception) as e:
            logger.error(e)
            session.rollback()
            raise e
        finally:
            session.close()

    return session_scope



from __future__ import annotations

from typing import Callable

from meiga import BoolResult

from petisco.base.domain.message.message import Message


class MessageHandlerReturnsNoneError(ValueError):
    def __init__(self, handler: Callable[[Message], BoolResult]):
        handler_name = getattr(handler, "__name__", repr(handler))
        handler_module = handler.__module__ + "."
        message = f"Subscribers (event handlers) must return a Result object. Please check {handler_module}{handler_name}"
        super().__init__(message)



from __future__ import annotations

from typing import TypeVar

from meiga import AnyResult

T = TypeVar("T")


def as_fastapi(result: AnyResult, expected_type: type[T] | None = None) -> T:
    """
    Transforms Controller Result to mapped value (Takes Config Controller configurations)
    """
    return result.transform(expected_type=expected_type)



import inspect
from abc import ABC
from typing import Any, Callable, Optional, Union

from meiga import AnyResult, Error, Result

from petisco.base.application.controller.error_map import ErrorMap
from petisco.base.application.controller.http_error import HttpError
from petisco.base.domain.errors.domain_error import DomainError


def default_failure_handler(result: AnyResult, error_map: ErrorMap) -> Union[HttpError, AnyResult]:
    error_type = type(result.value)
    mapped_result = error_map.get(error_type, result)
    return mapped_result


class ResultMapper(ABC):
    def __init__(
        self,
        error_map: Optional[ErrorMap] = None,
        success_handler: Callable[[AnyResult], Any] = lambda result: result,
        failure_handler: Callable[
            [Result[DomainError, Error], Union[ErrorMap, None]], Any
        ] = default_failure_handler,
    ):
        self.error_map = error_map if error_map is not None else {}
        self.success_handler = success_handler
        self.failure_handler = failure_handler

    def map(self, result: AnyResult) -> Any:
        try:
            if result.is_success:
                return self.success_handler(result)
            else:
                if "error_map" in inspect.signature(self.failure_handler).parameters:
                    return self.failure_handler(result, self.error_map)
                else:
                    return self.failure_handler(result)  # noqa
        except AttributeError as err:
            raise TypeError(
                f"Controller Error: Return value `{result}` ({type(result)}) must be a `meiga.Result` to "
                f"map values to success and failure handlers."
            ) from err



from __future__ import annotations

import os
import re
from typing import Any, Generic, TypeVar

from petisco.base.misc.builder import Builder

T = TypeVar("T")


class Dependency(Generic[T]):
    """
    Data structure to define how our dependency (Repository, AppService, etc) is built.

    Definition of default Builder is Mandatory, create alternative builder with envar_modifier is optional.
    """

    type: T
    builders: dict[str, Builder[T]]
    alias: str | None = None
    envar_modifier: str | None = None
    strict: bool = True

    def __init__(
        self,
        type: T,  # noqa
        *,
        builders: dict[str, Builder[Any]],
        alias: str | None = None,
        envar_modifier: str | None = None,
        strict: bool = True,
    ):
        self.type = type
        self.alias = alias
        self._set_envar_modifier(envar_modifier)
        self.builders = builders
        self.strict = strict
        self._check()

    def get_key(self) -> str:
        if self.alias:
            return f'{self.type.__name__} (alias="{self.alias}")'
        return self.type.__name__

    def _check(self) -> None:
        if self.builders is None or self.builders.get("default") is None:
            raise TypeError(
                f"Dependency: Define at least one builder for {self.get_key()} with `default` key. This is "
                f"mandatory. Given builders={self.builders} "
            )

    def _set_envar_modifier(self, envar_modifier: str | None = None):
        self.envar_modifier = None
        if envar_modifier is None and self.type:
            self.envar_modifier = (
                # "PETISCO_" + # TODO this will break compatibility (waiting for new versions)
                re.sub(r"(?<!^)(?=[A-Z])", "_", self.type.__name__).upper() + "_TYPE"
            )
            if self.alias:
                alias_str = re.sub(r"(?<!^)(?=[A-Z])", "_", self.alias).replace("-", "_").upper()
                self.envar_modifier = self.envar_modifier.replace("_TYPE", f"_ALIAS_{alias_str}_TYPE")
        else:
            self.envar_modifier = envar_modifier

    def _validate(self) -> None:
        if self.type and self.strict:
            if self.builders is None:
                return
            for key, builder in self.builders.items():
                klass = builder.klass
                if hasattr(builder.klass, "__origin__"):
                    klass = builder.klass.__origin__

                if not issubclass(klass, self.type):
                    raise TypeError(
                        f"Dependency: The class {builder.klass.__name__} from builders ({key}) is not a subclass from "
                        f"generic type given by Dependency[{self.type.__name__}]. If you want to skip it use "
                        f"Dependency(strict=False) "
                    )

    def get_instance(self) -> T:
        """
        Returns an instance of the dependency.
        """

        self._validate()

        modifier = os.getenv(self.envar_modifier) if self.envar_modifier else None
        if not modifier:
            builder = self._get_default_builder()
            instance = builder.build()
            return instance
        else:
            builder = self.builders.get(modifier)
            if not builder:
                builder = self._get_default_builder()
            assert isinstance(builder, Builder), "Oh no! Dependency builder is corrupted!"
            instance = builder.build()
            return instance

    def _get_default_builder(self) -> Builder:
        builder = self.builders.get("default")
        if not builder:
            raise TypeError(
                f"Dependency: Define a default builder for {self.type.__name__}. This is mandatory."
            )
        return builder



from petisco.extra.elastic_apm.is_elastic_apm_available import is_elastic_apm_available


def capture_exception() -> None:
    if is_elastic_apm_available():
        import elasticapm  # noqa

        client = elasticapm.get_client()
        if client:
            client.capture_exception()



from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.application.notifier.notifier_message import NotifierMessage


class NotImplementedNotifier(Notifier):
    def publish(self, message: NotifierMessage) -> None:
        pass

    def publish_exception(self, notifier_exception_message: NotifierExceptionMessage) -> None:
        pass



import contextlib
from typing import Any, Dict


class Singleton(type):
    _instances: Dict[Any, Any] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        force = kwargs.get("force_recreation", False)

        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        elif force:
            cls.clear()
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]

    def clear(cls) -> None:
        with contextlib.suppress(KeyError):
            del Singleton._instances[cls]



from random import random

from petisco.base.application.chaos.chaos_config import chaos_config


class ChaosInvalidMessagePublication(Exception): ...


def check_chaos_publication():
    value = random()
    if value < chaos_config.percentage_invalid_message_publication:
        raise ChaosInvalidMessagePublication()



from __future__ import annotations

import json

from loguru import logger
from redis.client import Redis
from redis.cluster import RedisCluster
from redis.exceptions import RedisError

from petisco.base.domain.errors.defaults.bus import BusCannotPublish
from petisco.base.domain.message.message import Message
from petisco.base.domain.message.message_bus import MessageBus, TypeMessage


class RedisMessageBus(MessageBus):
    """
    A generic implementation of MessageBus using Redis infrastructure.
    """

    def publish(self, message: TypeMessage | list[TypeMessage]) -> None:
        pass

    def __init__(
        self,
        organization: str,
        service: str,
        redis_database: Redis | RedisCluster,
        database_name: str,
    ):
        self.organization = organization
        self.service = service
        self.redis_database = redis_database
        self.database_name = database_name

    def save(self, messages: list[Message]) -> None:
        """
        Save several Message
        """

        updated_messages = []
        for message in messages:
            self._check_is_message(message)
            meta = self.get_configured_meta()
            message = message.update_meta(meta)
            updated_messages.append(message)
        try:
            data = [self._get_formatted_data(message) for message in updated_messages]
            with self.redis_database.pipeline() as pipe:
                pipe.lpush(self.database_name, *data)
                pipe.execute()
        except (TimeoutError, ConnectionError, RedisError) as exc:
            logger.opt(exception=True).error(
                f"Error publishing events ({len(messages)} of type {messages[0].get_message_type()}). Exception traceback:"
            )
            raise BusCannotPublish(additional_info={"error message": str(exc)}) from exc

    def _get_formatted_data(self, message: Message):
        formatted_data = {
            "organization": self.organization,
            "service": self.service,
            "message": message.format(),
        }
        return json.dumps(formatted_data)

    def close(self) -> None:
        pass



from typing import Any


def assert_http(response: Any, expected_status_code: int) -> None:
    response_detail = response.json if isinstance(response.json, dict) else response.json()
    assert response.status_code == expected_status_code, f"{response.status_code} - {response_detail}"



from typing import Union

from pydantic import Field, confloat
from pydantic_settings import BaseSettings


class ChaosConfig(BaseSettings):
    percentage_invalid_message_publication: confloat(ge=0.0, le=1.0) = Field(
        default=0,
        validation_alias="PETISCO_CHAOS_PERCENTAGE_INVALID_MESSAGE_PUBLICATION",
        description=(
            "Percentage of invalid message publication."
            " Where 1.0 means rejecting all the publishing messages.",
        ),
    )
    percentage_simulate_nack: confloat(ge=0.0, le=1.0) = Field(
        default=0,
        validation_alias="PETISCO_CHAOS_PERCENTAGE_SIMULATE_NACK",
        description="Percentage of simulate nack. Where 1.0 rejects all the consuming messages",
    )
    delay_after_message_consume_seconds: confloat(ge=0.0, le=60.0) = Field(
        default=0,
        validation_alias="PETISCO_CHAOS_DELAY_AFTER_MESSAGE_CONSUME_SECONDS",
        description="Delay message handler execution for a given number of seconds",
    )
    percentage_simulate_failures: confloat(ge=0.0, le=1.0) = Field(
        default=0,
        validation_alias="PETISCO_CHAOS_DELAY_AFTER_OPERATION_SECONDS",
        description="Percentage of simulate failures. Where 1.0 simulate always a failure on handlers",
    )
    protected_routing_keys: Union[str, None] = Field(
        default=None,
        validation_alias="PETISCO_CHAOS_PROTECTED_ROUTING_KEYS",
        description="Routing keys where chaos will not be applied",
    )


chaos_config = ChaosConfig()



import os
from functools import wraps
from inspect import signature
from typing import Any, Callable, Dict, List, Union

from loguru import logger
from meiga import Error, Failure, Result

from petisco.base.application.application_info import ApplicationInfo
from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.middleware.notifier_middleware import NotifierMiddleware
from petisco.base.application.middleware.print_middleware import PrintMiddleware
from petisco.base.domain.errors.critical_error import CriticalError
from petisco.base.domain.errors.unknown_error import UnknownError
from petisco.base.domain.value_objects.middleware_scope import MiddlewareScope
from petisco.base.misc.result_mapper import ResultMapper
from petisco.extra.elastic_apm.capture_exception import capture_exception
from petisco.extra.meiga import WaitingForEarlyReturn


def get_middleware_instances(config: Dict[str, Any]) -> List[Middleware]:
    middlewares_instances: List[Middleware] = []

    middlewares_configs = getattr(config, "middlewares", [])
    use_shared_middlewares = getattr(config, "use_shared_middlewares", True)
    if use_shared_middlewares:
        global_middlewares = get_global_middlewares()
        middlewares_configs = list(set(middlewares_configs + global_middlewares))

    for middlewares_config in middlewares_configs:
        if not isinstance(middlewares_config, Middleware):
            try:
                middlewares_instances.append(middlewares_config())
            except TypeError as exc:
                raise TypeError(
                    f"Middlewares cannot have configurable constructor, please review your petisco extension.\n{str(exc)}"
                ) from exc
        else:
            middlewares_instances.append(middlewares_config)

    return middlewares_instances


def update_middlewares(config: Dict[str, Any], middlewares: List[Middleware]) -> None:
    if config is not None and hasattr(config, "middlewares"):
        config.middlewares = middlewares


def get_global_middlewares() -> List[Middleware]:
    environment_middlewares = get_middlewares_configuration_from_environment()
    shared_middlewares = ApplicationInfo().shared_middlewares

    return list(set(environment_middlewares + shared_middlewares))


def get_middlewares_configuration_from_environment() -> List[Middleware]:
    def gettype(name: str) -> Middleware:
        lookup_table = {
            "NotifierMiddleware": NotifierMiddleware,
            "PrintMiddleware": PrintMiddleware,
        }
        my_type = lookup_table.get(name)
        if not my_type:
            raise TypeError(
                f"Value {name} in PETISCO_DEFAULT_MIDDLEWARES is not valid. Please, use {list(lookup_table.keys())}"
            )
        return my_type

    default_middlewares_names: Union[str, None] = os.getenv("PETISCO_DEFAULT_MIDDLEWARES")
    if not default_middlewares_names:
        return []
    default_middlewares_names_list = default_middlewares_names.split(",")
    return [gettype(middleware_name) for middleware_name in default_middlewares_names_list]


def wrapper(
    execute_func: Callable[..., Any],
    wrapped_class_name: str,
    config: Any,
    mapper: ResultMapper,
) -> Callable[..., Any]:
    @wraps(execute_func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        from petisco import Controller, MessageSubscriber

        middlewares = get_middleware_instances(config)
        arguments = signature(execute_func).bind(*args, **kwargs).arguments
        arguments.pop("self")

        for middleware in middlewares:
            if (
                middleware.scope == MiddlewareScope.ALL
                or middleware.scope == MiddlewareScope.CONTROLLER
                and issubclass(args[0].__class__, Controller)
                or middleware.scope == MiddlewareScope.SUBSCRIBER
                and issubclass(args[0].__class__, MessageSubscriber)
            ):
                pass
            else:
                continue

            try:
                middleware.set_data(wrapped_class_name, arguments)
                middleware.before()
            except Exception as exception:
                logger.error(f"Error in the {wrapped_class_name} middlewares (before).")
                logger.exception(exception)

        try:
            result = execute_func(*args, **kwargs)
        except WaitingForEarlyReturn as exc:
            result = exc.result
        except Error as error:
            result = Failure(error)
        except Exception as exception:
            unknown_error = UnknownError.from_exception(
                exception=exception,
                arguments=arguments if len(arguments) > 0 else args,
                class_name=wrapped_class_name,
            )
            result = Failure(unknown_error)
            capture_exception()

        if not isinstance(result, Result):
            raise TypeError(
                f"Controller Error: Return value `{result}` ({type(result)}) must be a `meiga.Result`"
            )

        if isinstance(result.value, CriticalError) and len(arguments) > 0:
            formatted_args = {
                k: str(v) if not isinstance(v, bytes) else "bytes" for k, v in arguments.items()
            }
            result.value.set_additional_info(formatted_args)

        result.set_transformer(mapper.map)
        for middleware in middlewares:
            if (
                middleware.scope == MiddlewareScope.ALL
                or middleware.scope == MiddlewareScope.CONTROLLER
                and issubclass(args[0].__class__, Controller)
                or middleware.scope == MiddlewareScope.SUBSCRIBER
                and issubclass(args[0].__class__, MessageSubscriber)
            ):
                pass
            else:
                continue

            if result:
                try:
                    middleware.after(result)
                except Exception as exception:
                    logger.error(f"Error in the {wrapped_class_name} middlewares (after).")
                    logger.exception(exception)

        update_middlewares(config, middlewares)
        return result

    return wrapped



from __future__ import annotations

from abc import ABC, abstractmethod
from inspect import iscoroutinefunction
from types import FunctionType
from typing import Any

from meiga import AnyResult, NotImplementedMethodError

from petisco.base.application.middleware.middleware import Middleware
from petisco.base.misc.async_wrapper import async_wrapper
from petisco.base.misc.result_mapper import ResultMapper
from petisco.base.misc.wrapper import wrapper


def get_mapper(bases: tuple[Any], config: dict[str, Any] | None) -> ResultMapper:
    mapper = ResultMapper()
    if config:
        for base in bases:
            method = getattr(base, "get_config_mapper", None)
            if method:
                mapper = method(config)

    else:
        for base in bases:
            method = getattr(base, "get_default_mapper", None)
            if method:
                mapper = method()

    return mapper


class MetaController(type, ABC):
    middlewares: list[Middleware] = []
    use_global_middlewares: bool = True

    def __new__(mcs, name: str, bases: tuple[Any], namespace: dict[str, Any]) -> MetaController:
        config = namespace.get("Config")

        mapper = get_mapper(bases, config)

        if "execute" not in namespace:
            raise NotImplementedError("Petisco Controller must implement an execute method")

        new_namespace = {}
        for attributeName, attribute in namespace.items():
            if isinstance(attribute, FunctionType) and attribute.__name__ == "execute":
                if iscoroutinefunction(attribute):
                    attribute = async_wrapper(attribute, name, config, mapper)
                else:
                    attribute = wrapper(attribute, name, config, mapper)
            new_namespace[attributeName] = attribute

        return super().__new__(mcs, name, bases, new_namespace)

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError



from __future__ import annotations

from redis.client import Redis
from redis.cluster import RedisCluster

from petisco.base.domain.message.command import Command
from petisco.extra.redis.application.message.bus.redis_message_bus import (
    RedisMessageBus,
)


class RedisCommandBus(RedisMessageBus):
    """
    An implementation of CommandBus using Redis infrastructure.
    """

    def __init__(self, organization: str, service: str, redis_database: Redis | RedisCluster):
        super().__init__(organization, service, redis_database, "commands")

    def dispatch(self, command: Command | list[Command]) -> None:
        """
        Dispatch one Command
        """
        commands = self._check_input(command)
        self.save(commands)



from petisco.extra.logger.log_message import LogMessage
from petisco.extra.logger.logger import (
    CRITICAL,
    DEBUG,
    ERROR,
    FATAL,
    INFO,
    NOTSET,
    WARN,
    WARNING,
    Logger,
)
from petisco.extra.logger.logging_based_logger import LoggingBasedLogger
from petisco.extra.logger.loguru_logger import LoguruLogger

classes = ["Logger", "LogMessage", "LoggingBasedLogger", "LoguruLogger"]
constants = ["CRITICAL", "FATAL", "ERROR", "WARNING", "WARN", "INFO", "DEBUG", "NOTSET"]

__all__ = classes + constants



from abc import ABC
from copy import copy
from inspect import isclass
from typing import Any, Dict, List, Union, get_args

from pydantic import (
    BaseModel,
    Field,
    NonNegativeInt,
    PrivateAttr,
    field_serializer,
    field_validator,
    model_validator,
)

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.model.uuid import Uuid
from petisco.base.domain.model.value_object import ValueObject

DEFAULT_VERSION = 1


class AggregateRoot(ABC, BaseModel):
    """
    A base class to define AggregateRoot

    It is a cluster of associated entities that are treated as a unit for the purpose of data changes.
    """

    aggregate_id: Uuid = Field(default_factory=Uuid.v4)
    aggregate_version: NonNegativeInt = Field(default=DEFAULT_VERSION)
    _domain_events: List[DomainEvent] = PrivateAttr(default=[])

    @model_validator(mode="before")
    def model_validation(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        new_data = copy(data)
        for key, annotation in cls.__annotations__.items():
            value = data.get(key)
            if value is None:
                continue

            union_annotations = get_args(annotation)
            if len(union_annotations) > 0:
                for union_annotation in union_annotations:
                    if (
                        isclass(union_annotation)
                        and issubclass(union_annotation, ValueObject)
                        and isinstance(value, str)
                    ):
                        new_value = union_annotation(value=value)
                        new_data[key] = new_value
            else:
                if isclass(annotation) and issubclass(annotation, ValueObject) and isinstance(value, str):
                    new_value = annotation(value=value)
                    new_data[key] = new_value
        return new_data

    @field_serializer("aggregate_id")
    def serialize_aggregate_id(self, aggregate_id: Uuid) -> str:
        return aggregate_id.value

    @field_validator("aggregate_id", mode="before")
    def set_aggregate_id(cls, v: Union[str, Uuid]) -> Uuid:
        v = Uuid(value=v) if isinstance(v, str) else v
        return v or Uuid.v4()

    @field_validator("aggregate_version", mode="before")
    def set_aggregate_version(cls, v: NonNegativeInt) -> NonNegativeInt:
        return v or DEFAULT_VERSION

    def record(self, domain_event: DomainEvent) -> None:
        """
        Record something that happened is our domain related with the aggregate (a DomainEvent).
        """
        self._domain_events.append(domain_event)

    def clear_domain_events(self) -> None:
        """
        Clear all domain events recorded.
        """
        self._domain_events = []

    def pull_domain_events(self) -> List[DomainEvent]:
        """
        Returns and clear all domain events recorded.
        """
        domain_events = copy(self._domain_events)
        self.clear_domain_events()
        return domain_events

    def get_domain_events(self) -> List[DomainEvent]:
        """
        Returns all domain events recorded.
        """
        return self._domain_events

    def get_first_domain_event(self) -> Union[DomainEvent, None]:
        """
        Returns only the first domain event recorded.
        """
        return self._domain_events[0] if len(self._domain_events) > 0 else None

    def get_last_domain_event(self) -> Union[DomainEvent, None]:
        """
        Returns only the last domain event recorded.
        """
        return self._domain_events[-1] if len(self._domain_events) > 0 else None



from __future__ import annotations

import os

from loguru import logger

from petisco.base.application.application_configurer import ApplicationConfigurer
from petisco.base.application.dependency_injection.container import Container
from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.domain.errors.unknown_error import UnknownError
from petisco.base.domain.message.command_bus import CommandBus
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.base.domain.message.message_configurer import MessageConfigurer
from petisco.base.domain.message.message_consumer import MessageConsumer
from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_message_consumer import (
    RabbitMqMessageConsumer,
)

MAX_RETRIES = 5

CLEAR_SUBSCRIBER_BEFORE: bool = os.getenv(
    "PETISCO_RABBITMQ_CONFIGURER_CLEAR_SUBSCRIBER_BEFORE", "false"
).lower() in ["true", "1", "yes", "on"]
CLEAR_STORE_BEFORE: bool = os.getenv("PETISCO_RABBITMQ_CONFIGURER_CLEAR_STORE_BEFORE", "false").lower() in [
    "true",
    "1",
    "yes",
    "on",
]


class RabbitMqConfigurer(ApplicationConfigurer):
    def __init__(
        self,
        subscribers: list[type[MessageSubscriber]],
        execute_after_dependencies: bool = True,
        start_consuming: bool = True,
        alias: str | None = None,
        inner_bus_organization: str | None = None,
        inner_bus_service: str | None = None,
        clear_subscriber_before: bool | None = None,
        clear_store_before: bool | None = None,
        use_container_buses: bool = True,
    ):
        """
        Initializes an instance of RabbitMqConfigurer.

        Args:
            subscribers (list[MessageSubscriber]): A list of MessageSubscriber objects representing
                the message subscribers.
            execute_after_dependencies (bool, optional): Flag indicating whether to execute after
                dependencies. Defaults to True.
            start_consuming (bool, optional): Flag indicating whether to start consuming messages.
                Defaults to True.
            alias (str, optional): Alias for the MessageConfigurer and MessageConsumer
                (Container.get(MessageConfigurer|MessageConsumer, alias=self.alias)). Defaults to None.
            inner_bus_organization (str, optional): configured organization for inner buses (DomainEventBus
                and CommandBus). If None will configure consumer organization.
            inner_bus_service (str, optional): configured service for inner buses (DomainEventBus
                and CommandBus). If None will configure consumer service.
            use_container_buses (bool, optional): configure buses from petisco.Container.
        """
        self.subscribers = subscribers
        self.start_consuming = start_consuming
        self.alias = alias
        self.inner_bus_organization = inner_bus_organization
        self.inner_bus_service = inner_bus_service
        self.consumer: MessageConsumer | None = None
        self.clear_subscriber_before = (
            clear_subscriber_before if clear_subscriber_before else CLEAR_SUBSCRIBER_BEFORE
        )
        self.clear_store_before = clear_store_before if clear_store_before else CLEAR_STORE_BEFORE
        self.use_container_buses = use_container_buses

        super().__init__(execute_after_dependencies)

    def execute(self, testing: bool = False) -> None:
        configurer = Container.get(MessageConfigurer, alias=self.alias)
        try:
            configurer.configure_subscribers(
                self.subscribers,
                clear_subscriber_before=self.clear_subscriber_before,
                clear_store_before=self.clear_store_before,
            )

            if self.start_consuming:
                self.consumer = Container.get(MessageConsumer, alias=self.alias)

                if isinstance(self.consumer, RabbitMqMessageConsumer):
                    self.consumer.set_inner_bus_config(self.inner_bus_organization, self.inner_bus_service)
                    if self.use_container_buses is True:

                        def domain_event_bus_builder() -> DomainEventBus:
                            return Container.get(DomainEventBus)

                        def command_bus_builder() -> CommandBus:
                            return Container.get(CommandBus)

                        self.consumer.domain_event_bus_builder = domain_event_bus_builder
                        self.consumer.command_bus_builder = command_bus_builder

                self.consumer.add_subscribers(self.subscribers)
                self.consumer.start()
        except ConnectionError as ex:
            logger.error(f"Connection error with rabbit when trying to configure it. Message {str(ex)}")
            self.notify_connection_error(ex)

    def stop(self) -> None:
        if self.consumer:
            self.consumer.stop()

    def notify_connection_error(self, exception: Exception) -> None:
        notifier = Container.get(Notifier)
        error = UnknownError.from_exception(
            exception=exception,
            arguments={
                "service": self.inner_bus_service,
                "subscribers": self.subscribers,
            },
        )
        notifier_exception_message = NotifierExceptionMessage.from_unknown_error(
            error, title="Connection error trying to configure RabbitMQ"
        )
        notifier.publish_exception(notifier_exception_message)



from __future__ import annotations

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus


class NotImplementedDomainEventBus(DomainEventBus):
    def publish(self, domain_event: DomainEvent | list[DomainEvent]) -> None:
        domain_events = [domain_event] if isinstance(domain_event, DomainEvent) else domain_event
        for domain_event in domain_events:
            self._check_is_domain_event(domain_event)
            meta = self.get_configured_meta()
            _ = domain_event.update_meta(meta)

    def retry_publish(
        self,
        domain_event: DomainEvent,
        retry_routing_key: str,
        retry_exchange_name: str | None = None,
    ) -> None:
        self._check_is_domain_event(domain_event)
        meta = self.get_configured_meta()
        _ = domain_event.update_meta(meta)

    def close(self) -> None:
        pass



from abc import ABC, abstractmethod
from typing import Any, Dict, Union

from loguru import logger
from meiga import AnyResult

from petisco.base.domain.value_objects.middleware_scope import MiddlewareScope


class Middleware(ABC):
    """
    A base class to model your Middleware.
    A Middleware works before some operations (Controller and Subscriber) and also before returning the final result.
    """

    wrapped_class_name: Union[str, None] = None
    wrapped_class_input_arguments: Union[str, None] = None
    scope: MiddlewareScope = MiddlewareScope.ALL

    def set_data(
        self,
        wrapped_class_name: str,
        wrapped_class_input_arguments: Any,
        scope: MiddlewareScope = MiddlewareScope.ALL,
    ) -> None:
        self.wrapped_class_name = wrapped_class_name
        self.wrapped_class_input_arguments = wrapped_class_input_arguments
        self.scope = scope

    @abstractmethod
    def before(self) -> None:
        pass

    @abstractmethod
    def after(self, result: AnyResult) -> None:
        pass

    def get_meta_from_input(self) -> Dict[str, Any]:
        # This method get info from legacy info_id model to keep compatibility
        meta = {}

        try:
            info_id = self.wrapped_class_input_arguments.get("info_id")
            if info_id and hasattr(info_id, "to_meta"):
                meta = info_id.to_meta().get("info_id", {})
        except Exception as exc:
            logger.error(f"Middleware error getting info_id on get_meta_from_input: {str(exc)}")

        return meta



from typing import Any, Dict, Union

from pydantic import BaseModel, Field, field_validator

DEFAULT_HTTP_ERROR_DETAIL = "Unknown Error"


class HttpError(BaseModel):
    status_code: int = Field(default=500)
    detail: str = Field(default=DEFAULT_HTTP_ERROR_DETAIL)
    headers: Union[Dict[str, Any], None] = Field(default=None)
    type_error: Union[str, None] = Field(default=None, validate_default=True)

    @field_validator("type_error", mode="before")
    def prevent_type_error_none(cls, value: Union[str, None]) -> str:
        if value is None:
            value = cls.__name__
        return value



from fastapi import FastAPI
from fastapi.routing import APIRoute
from loguru import logger


def is_async_callable(route: APIRoute) -> bool:
    endpoint = route.endpoint
    import asyncio

    return bool(asyncio.iscoroutinefunction(endpoint))


def ensure_all_routers_are_async(app: FastAPI) -> None:
    for route in app.routes:
        if isinstance(route, APIRoute) and not is_async_callable(route):
            logger.error(f"Router with {route.path} is not using async definition")
            raise SystemError(f"Router of {route.path} is not using async definition")
